/*
   ImageManager - MyIDE Image Manager information

   Copyright (C) 2007 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string.h>
#include "ImageManager.h"
#include "Error.h"
#include "AtariDebug.h"
#include "SimpleProgressIndicator.h"

ImageManager::ImageManager()
	: fInitialized(false),
	  fChanged(false)
{
}

ImageManager::~ImageManager()
{
}

ImageManager::TableEntry::TableEntry()
	: CachedName(0)
{
}

ImageManager::TableEntry::TableEntry(const ImageManager::TableEntry& other)
	: SlotNumber(other.SlotNumber),
	  Status(other.Status),
	  Category(other.Category),
	  CachedName(0)
{
	if (other.CachedName) {
		CachedName = strdup(other.CachedName);
	}
}

ImageManager::TableEntry::~TableEntry()
{
	if (CachedName) {
		free(CachedName);
		CachedName = 0;
	}
}

ImageManager::TableEntry& ImageManager::TableEntry::operator=(const ImageManager::TableEntry& other)
{
	if (CachedName) {
		free(CachedName);
	}
	CachedName = 0;
	SlotNumber = other.SlotNumber;
	Status = other.Status;
	Category = other.Category;
	if (other.CachedName) {
		CachedName = strdup(other.CachedName);
	}
	return *this;
}

void ImageManager::InitBlank()
{
	memset(fConfig, 0, 128);

	fConfig[eOfsInitFlag] = 1;
	fConfig[eOfsNumberOfEntries] = 0;
	fConfig[eOfsNumberOfEntries+1] = 0;
	fConfig[eOfsOptionByte] = 0x81;

	fTableEntries.clear();

	int i;
	const char *catstr;
	for (i=0;i<20;i++) {
		switch (i) {
		case 0: catstr="game"; break;
		case 1: catstr="demo"; break;
		case 2: catstr="util"; break;
		case 3: catstr="dos"; break;
		case 4: catstr="misc"; break;
		case 5: catstr="gfx"; break;
		case 6: catstr="musx"; break;
		case 7: catstr="prog"; break;
		case 8: catstr="dev"; break;
		default: catstr=""; break;
		}
		strncpy((char*)fConfig + eOfsCategoryInfo + i*eCategoryLength, catstr, eCategoryLength);
	}
	fConfig[eOfsSearchCase] = 0;
	fConfig[eOfsGetApeDesc] = 0;

	fInitialized = true;
	fChanged = true;

	if (!ParseConfigBlock()) {
		DPRINTF("internal error: parsing newly initialized config block failed\n");
		Assert(false);
	}

	for (i=0; i<eMaxEntries; i++) {
		fSlotToTableMap[i] = eUnusedSlotNumber;
	}
}

bool ImageManager::ParseConfigBlock()
{
	if (fConfig[eOfsInitFlag] != 1) {
		DPRINTF("not a valid Image Manager config\n");
		return false;
	}
	fNumberOfEntries = fConfig[eOfsNumberOfEntries] + (fConfig[eOfsNumberOfEntries + 1] << 8);
	if (fNumberOfEntries >= eMaxEntries) {
		DPRINTF("illegal number of entries (%d) in Image Manager config\n", fNumberOfEntries);
		return false;
	}
	unsigned int i, j, k, ofs;
	unsigned char c;
	for (i=0;i<20;i++) {
		for (j=0;j<eCategoryLength - 1;j++) {
			fCategories[i][j] = ' ';
		}
		fCategories[i][eCategoryLength - 1] = 0;
		ofs = eOfsCategoryInfo + i * eCategoryLength;
		j=0;
		k=0;
		while (j < eCategoryLength && (c = fConfig[ofs + j]) != 0) {
			c &= 0x7f;
			if (c >= 32 && c <= 124 && c != 96 && c != 123 ) {
				fCategories[i][k++] = c;
			}
			j++;
		}
	}
	return true;
}

bool ImageManager::ReadFromDisk(RCPtr<MyideDisk>& disk)
{
	fInitialized = false;
	fChanged = false;

	fTableEntries.clear();

	if (!disk->ImageSpaceEnabled()) {
		DPRINTF("error: image space is not enabled\n");
		return false;
	}
	unsigned int imageno = disk->GetNumberOfImageSlots();

	if (!disk->ReadImageSector(imageno, 1, eSD, fConfig)) {
		DPRINTF("error reading Image Manager config\n");
		return false;
	}
	if (!ParseConfigBlock()) {
		return false;
	}
	unsigned int i;
	for (i=0; i<eMaxEntries; i++) {
		fSlotToTableMap[i] = eUnusedSlotNumber;
	}

	unsigned char buf[256];
	unsigned int ofs;
	unsigned int slot;
	TableEntry entry;
	for (i=0; i<fNumberOfEntries; i++) {
		ofs = (i & 0x3f) * 4;
		if (ofs == 0) {
			if (!disk->ReadImageSector(imageno, 4 + (i >> 6), eDD, buf)) {
				DPRINTF("error reading Image Manager table\n");
				return false;
			}
		}
		slot = buf[ofs] + (buf[ofs+1] << 8);
		if (slot >= fNumberOfEntries) {
			DPRINTF("illegal slot number (%d) in Image Manager table\n", slot+1);
			return false;
		}
		if (fSlotToTableMap[slot] != eUnusedSlotNumber) {
			DPRINTF("slot %d found twice in Image Manager table\n", slot+1);
			return false;
		}
		fSlotToTableMap[slot] = i;
		entry.SlotNumber = slot;
		entry.Status = buf[ofs+2];
		if (buf[ofs+3] >= eNumCategories) {
			DPRINTF("illegal category number (%d) in Image manager table\n", buf[ofs+3]+1);
			return false;
		}
		entry.Category = buf[ofs+3];
		fTableEntries.push_back(entry);
	}
	Assert(fTableEntries.size() == fNumberOfEntries);

	fInitialized = true;
	fChanged = false;
	return true;
}

bool ImageManager::WriteToDisk(RCPtr<MyideDisk>& disk) const
{
	if (!IsInitialized()) {
		DPRINTF("error: trying to write non initialized Image Manager data to disk\n");
		return false;
	}
	if (!IsChanged()) {
		DPRINTF("warning: Image Manager unchanged, writing it to disk anyway\n");
	}

	if (!disk->ImageSpaceEnabled()) {
		DPRINTF("error: image space is not enabled\n");
		return false;
	}
	unsigned int imageno = disk->GetNumberOfImageSlots();

	if (!disk->WriteImageSector(imageno, 1, eSD, fConfig)) {
		DPRINTF("error writing Image Manager config\n");
		return false;
	}

	unsigned int i;
	unsigned char buf[256];
	memset(buf, 0, 256);
	unsigned int ofs;

	for (i=0; i<fNumberOfEntries; i++) {
		ofs = (i & 0x3f) * 4;
		buf[ofs] = fTableEntries[i].SlotNumber & 0xff;
		buf[ofs+1] = fTableEntries[i].SlotNumber >> 8;
		buf[ofs+2] = fTableEntries[i].Status;
		buf[ofs+3] = fTableEntries[i].Category;

		if ((i + 1 == fNumberOfEntries) || ((i & 0x3f) == 0x3f)) {
			if (!disk->WriteImageSector(imageno, 4 + (i >> 6), eDD, buf)) {
				DPRINTF("error writing Image Manager table\n");
				return false;
			}
			memset(buf, 0, 256);
		}
	}
	MyideImage img;
	img.Density = eDD;
	sprintf(img.Name, "MyIDE Image Manager Config");
	if (!disk->SetImageInfo(imageno, img)) {
		DPRINTF("error setting image info of Image Manager config\n");
		return false;
	}
	fChanged = false;
	return true;
}

void ImageManager::SetNumberOfEntries(unsigned int entries)
{
	if (entries >= eMaxEntries) {
		DPRINTF("internal error: illegal number of entries in Image Manager config\n");
		Assert(false);
	}
	fNumberOfEntries = entries;
	fConfig[eOfsNumberOfEntries] = entries & 0xff;
	fConfig[eOfsNumberOfEntries+1] = entries >> 8;
	fChanged = true;
}

unsigned int ImageManager::GetSlotNumberOfEntry(unsigned int entry) const
{
	if (!IsValidEntryNumber(entry)) {
		return 0;
	}
	return fTableEntries[entry-1].SlotNumber + 1;
}

unsigned char ImageManager::GetStatusOfEntry(unsigned int entry) const
{
	if (!IsValidEntryNumber(entry)) {
		return 0;
	}
	return fTableEntries[entry-1].Status;
}

const char* ImageManager::GetCategoryOfEntry(unsigned int entry) const
{
	if (!IsValidEntryNumber(entry)) {
		return 0;
	}
	return fCategories[fTableEntries[entry-1].Category];
}

bool ImageManager::SetImageInformation(unsigned int slotNumber, unsigned char status, unsigned char category)
{
	if (slotNumber == 0) {
		DPRINTF("illegal slot number 0\n");
		return false;
	}
	slotNumber--;
	if (slotNumber > fNumberOfEntries) {
		DPRINTF("cannot add slot non-consecutively to Image Manager\n");
		return false;
	}

	if (fNumberOfEntries >= eMaxEntries) {
		DPRINTF("image manager table full\n");
		return false;
	}

	if (category > 20) {
		DPRINTF("illegal category number %d\n", category);
		return false;
	}

	if (slotNumber < fNumberOfEntries) {
		// update mode
		unsigned int idx = fSlotToTableMap[slotNumber];

		if (idx == eUnusedSlotNumber) {
			DPRINTF("internal error in image manager table\n");
			return false;
		}
		if (category != 0) {
			fTableEntries[idx].Category = category - 1;
		}
		fTableEntries[idx].Status = status;
		if (fTableEntries[idx].CachedName) {
			free(fTableEntries[idx].CachedName);
		}
		fTableEntries[idx].CachedName = 0;
	} else {
		// append mode
		if (fSlotToTableMap[slotNumber] != eUnusedSlotNumber) {
			DPRINTF("internal error in image manager table\n");
			return false;
		}
		if (category == 0) {
			category = 1;
		}
		TableEntry entry;
		entry.SlotNumber = slotNumber;
		entry.Status = status;
		entry.Category = category - 1;
		fTableEntries.push_back(entry);
		SetNumberOfEntries(fNumberOfEntries + 1);

		Assert(fTableEntries.size() == fNumberOfEntries);
	}
	fChanged = true;
	return true;
}

unsigned char ImageManager::DensityToStatusCode(EDensity density) const
{
	unsigned char status;
	switch (density) {
	case eSD: status = 0; break;
	case eED: status = ImageManager::eBit1040SectorsDisk; break;
	case eDD: status = ImageManager::eBitDoubleDensity; break;
	default:  status = ImageManager::eBit1040SectorsDisk | ImageManager::eBitDoubleDensity; break;
	}
	return status;
}

bool ImageManager::SetImageInformation(unsigned int slotNumber, EDensity density, bool enableBasic, unsigned char category)
{
	unsigned char status = DensityToStatusCode(density);
	if (enableBasic) {
		status |= ImageManager::eBitEnableBasic;
	}
	return SetImageInformation(slotNumber, status, category);
}

void ImageManager::RebuildSlotIndex()
{
	unsigned int i;
	for (i=0; i<fNumberOfEntries; i++) {
		fSlotToTableMap[fTableEntries[i].SlotNumber] = i;
	}
}

static bool CompareNameAscending(const ImageManager::TableEntry& e1, const ImageManager::TableEntry& e2)
{
	return (strcasecmp(e1.CachedName, e2.CachedName) < 0);
}

static bool CompareNameDescending(const ImageManager::TableEntry& e1, const ImageManager::TableEntry& e2)
{
	return (strcasecmp(e1.CachedName, e2.CachedName) > 0);
}

static bool CompareCategoryAscending(const ImageManager::TableEntry& e1, const ImageManager::TableEntry& e2)
{
	return (e1.Category < e2.Category);
}

static bool CompareCategoryDescending(const ImageManager::TableEntry& e1, const ImageManager::TableEntry& e2)
{
	return (e1.Category > e2.Category);
}

static bool CompareSlotNumber(const ImageManager::TableEntry& e1, const ImageManager::TableEntry& e2)
{
	return (e1.SlotNumber < e2.SlotNumber);
}

bool ImageManager::SortImageTable(RCPtr<MyideDisk>& disk, ESortMode sortMode)
{
	if (fNumberOfEntries < 2) {
		return true;
	}

	if (sortMode == eSortNameAscending || sortMode == eSortNameDescending) {
		RCPtr<SimpleProgressIndicator> progress = new SimpleProgressIndicator(fNumberOfEntries);
		printf("caching image names: ");
		progress->DisplayProgress(0);
		unsigned int i;
		MyideImage img;
		for (i=0; i<fNumberOfEntries; i++) {
			if (fTableEntries[i].CachedName == 0) {
				if (!disk->GetImageInfo(fTableEntries[i].SlotNumber + 1, img)) {
					printf("error reading image information of slot %d\n", fTableEntries[i].SlotNumber + 1);
					return false;
				}
				fTableEntries[i].CachedName = strdup(img.Name);
			}
			progress->DisplayProgress(i+1);
		}
		progress->FinishProgress();
		printf("\n");
	}
	switch (sortMode) {
	case eSortSlotNumber:
		printf("sorting image manager table by slot number: "); fflush(stdout);
		std::stable_sort(fTableEntries.begin(), fTableEntries.end(), CompareSlotNumber); break;
	case eSortNameAscending:
		printf("sorting image manager table by name: "); fflush(stdout);
		std::stable_sort(fTableEntries.begin(), fTableEntries.end(), CompareNameAscending); break;
	case eSortNameDescending:
		printf("sorting image manager table by name, descending: "); fflush(stdout);
		std::stable_sort(fTableEntries.begin(), fTableEntries.end(), CompareNameDescending); break;
	case eSortCategoryAscending:
		printf("sorting image manager table by category: "); fflush(stdout);
		std::stable_sort(fTableEntries.begin(), fTableEntries.end(), CompareCategoryAscending); break;
	case eSortCategoryDescending:
		printf("sorting image manager table by category, descending: "); fflush(stdout);
		std::stable_sort(fTableEntries.begin(), fTableEntries.end(), CompareCategoryDescending); break;
	default:
		assert(false);
	}
	printf("OK\n");
	fChanged = true;
	RebuildSlotIndex();
	return true;
}
