#ifndef IMAGEMANAGER_H
#define IMAGEMANAGER_H

/*
   ImageManager - MyIDE Image Manager information

   Copyright (C) 2007 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <vector>
#include <algorithm>
#include "RefCounted.h"
#include "RCPtr.h"
#include "MyideDisk.h"
#include "AtariDebug.h"

class ImageManager : public RefCounted {
public:
	ImageManager();
	virtual ~ImageManager();

	enum { eNumCategories = 20 };
	enum { eMaxEntries = 717*64 };

	void InitBlank();

	bool ReadFromDisk(RCPtr<MyideDisk>& disk);
	bool WriteToDisk(RCPtr<MyideDisk>& disk) const;

	inline bool IsInitialized() const;
	inline bool IsChanged() const;

	inline unsigned int GetNumberOfEntries() const;

	inline const char* GetCategoryName(unsigned int category) const;

	enum EStatusBits {
		eBitEnableBasic = 0x10,
		eBit1040SectorsDisk = 0x20,
		eBitDoubleDensity = 0x40,
		eBitDeleted = 0x80
	};

	// access table infos
	unsigned int GetSlotNumberOfEntry(unsigned int entry) const;
	unsigned char GetStatusOfEntry(unsigned int entry) const;
	const char* GetCategoryOfEntry(unsigned int entry) const;

	bool SetImageInformation(unsigned int slotNumber, unsigned char status, unsigned char category);
	bool SetImageInformation(unsigned int slotNumber, EDensity density, bool enableBasic, unsigned char category);

	unsigned char DensityToStatusCode(EDensity density) const;

	enum ESortMode {
		eSortSlotNumber,
		eSortNameAscending,
		eSortNameDescending,
		eSortCategoryAscending,
		eSortCategoryDescending
	};

	// sort the image table
	bool SortImageTable(RCPtr<MyideDisk>& disk, ESortMode sortMode);

	class TableEntry {
	public:
		TableEntry();
		TableEntry(const TableEntry& other);
		~TableEntry();
		TableEntry& operator=(const TableEntry& other);
		unsigned int SlotNumber;
		unsigned char Status;
		unsigned char Category;
		char* CachedName;
	};

private:
	inline bool IsValidEntryNumber(unsigned int entry) const;

	bool ParseConfigBlock();
	void SetNumberOfEntries(unsigned int entries);

	void RebuildSlotIndex();

	// raw configuration data
	unsigned char fConfig[128];

	enum EConfigOffsets {
		eOfsInitFlag = 0,
		eOfsNumberOfEntries = 1,
		eOfsOptionByte = 3,
		eOfsCategoryInfo = 4,
		eOfsSearchCase = 104,
		eOfsGetApeDesc = 105
	};

	// extraced configuration data
	unsigned int fNumberOfEntries;
	enum { eCategoryLength = 5 };
	char fCategories[eNumCategories][eCategoryLength];

	enum { eUnusedSlotNumber = 65535 };
	std::vector<struct TableEntry> fTableEntries;
	unsigned int fSlotToTableMap[eMaxEntries];

	bool fInitialized;
	mutable bool fChanged;
};

inline bool ImageManager::IsInitialized() const
{
	return fInitialized;
}

inline bool ImageManager::IsChanged() const
{
	return fChanged;
}

inline unsigned int ImageManager::GetNumberOfEntries() const
{
	Assert(IsInitialized());
	return fNumberOfEntries;
}

inline bool ImageManager::IsValidEntryNumber(unsigned int entry) const
{
	Assert (IsInitialized());
	return (entry >= 1) && (entry <= fNumberOfEntries);
}

inline const char* ImageManager::GetCategoryName(unsigned int category) const
{
	Assert (IsInitialized());
	if (category < 1 || category > eNumCategories) {
		return 0;
	} else {
		return fCategories[category - 1];
	}
}

#endif
