/*
   MyidePartitionInfoCHS - infos for (old) CHS partition table

   Copyright (C) 2008 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string.h>

#include "MyidePartitionInfoCHS.h"
#include "MiscUtils.h"

MyidePartitionInfoCHS::MyidePartitionInfoCHS()
{ }

MyidePartitionInfoCHS::~MyidePartitionInfoCHS()
{ }

bool MyidePartitionInfoCHS::ParsePartitionInfo(const unsigned char* buf)
{
	fCylinders = buf[0] + 256 * buf[1] + 1;
	fHeads = buf[2] + 1;
	fSectors = buf[3];

	fSectorsPerCylinder = fHeads * fSectors;

	fNumberOfPartitions = buf[4];

	fDriveBits = buf[5];
	fOptionByte = buf[14];

	fImageSpaceCylinder = 256 * buf[6];
	fCylindersPerImageSlot = buf[7];

	for (int i=0;i<eMaxPartitions;i++) {
		int idx = i*8+8;
		fPartitionInfo[i].DriveNumber = buf[idx];
		fPartitionInfo[i].StartingCylinder = buf[idx+1] + 256 * buf[idx+2];
		fPartitionInfo[i].EndingCylinder = buf[idx+3] + 256 * buf[idx+4];
		fPartitionInfo[i].MaxSectors =
			(fPartitionInfo[i].EndingCylinder - fPartitionInfo[i].StartingCylinder) *
			fSectorsPerCylinder;
		if (fPartitionInfo[i].MaxSectors > 65535) {
			fPartitionInfo[i].MaxSectors = 65535;
		}
		if (buf[idx+5] == 1) {
			fPartitionInfo[i].Density = eDD;
		} else {
			fPartitionInfo[i].Density = eSD;
		}
		if (fPartitionInfo[i].DriveNumber >= 1 && fPartitionInfo[i].DriveNumber <=8) {
			fPartitionInfo[i].IsActive = DriveIncludedInDriveBits(fPartitionInfo[i].DriveNumber);
		} else {
			fPartitionInfo[i].IsActive = false;
		}
	}
	return true;
}

bool MyidePartitionInfoCHS::CreatePartitionInfo(unsigned char* buf) const
{
	memset(buf, 0, 256);

	buf[0] = (fCylinders - 1) & 0xff;
	buf[1] = (fCylinders - 1) >> 8;
	buf[2] = fHeads - 1;
	buf[3] = fSectors;

	buf[4] = fNumberOfPartitions;

	buf[5] = fDriveBits;

	buf[6] = fImageSpaceCylinder >> 8;
	buf[7] = fCylindersPerImageSlot;

	buf[14] = fOptionByte;

	for (int i=0;i<eMaxPartitions;i++) {
		int idx = i*8+8;
		buf[idx] = fPartitionInfo[i].DriveNumber;
		buf[idx+1] = fPartitionInfo[i].StartingCylinder & 0xff;
		buf[idx+2] = fPartitionInfo[i].StartingCylinder >> 8;
		buf[idx+3] = fPartitionInfo[i].EndingCylinder & 0xff;
		buf[idx+4] = fPartitionInfo[i].EndingCylinder >> 8;
		if (fPartitionInfo[i].Density == eDD) {
			buf[idx+5] = 1;
		} else {
			buf[idx+5] = 0;
		}
	}
	return true;
}

bool MyidePartitionInfoCHS::IsChsPartitionTable() const
{
	return true;
}

void MyidePartitionInfoCHS::PrintPartitionTable() const
{
	printf("\nMyIDE 3.x/4.x CHS partition table:\n");
	printf("%d cylinders %d heads %d sectors (%d sectors per cyl)\n",
		fCylinders, fHeads, fSectors,
		fHeads*fSectors);
        printf("Image space: ");
        if (ImageSpaceEnabled()) {
                printf("first cylinder: %d  cylinders per slot: %d  total slots: %d\n",
			fImageSpaceCylinder,
			fCylindersPerImageSlot,
                        GetNumberOfImageSlots());
        } else {
                printf("disabled\n");
        }
        printf("partitions: %d", fNumberOfPartitions);
	char bin[9];
	MiscUtils::byte_to_bin(fOptionByte, bin);
        printf(" option byte: %s", bin);
	MiscUtils::byte_to_bin(fDriveBits, bin);
        printf(" drive bits: %s\n", bin);

        printf("part active drv start   end dns sectors\n");
        for (int i=1;i<=eMaxPartitions;i++) {
                printf("%d    %s    D%d: %5d %5d %s    %5d\n", i,
                        fPartitionInfo[i-1].IsActive ? "yes" : "no ",
                        fPartitionInfo[i-1].DriveNumber,
			fPartitionInfo[i-1].StartingCylinder,
			fPartitionInfo[i-1].EndingCylinder,
                        fPartitionInfo[i-1].Density == eDD ? "DD" : "SD",
                        fPartitionInfo[i-1].MaxSectors);
        }

}

void MyidePartitionInfoCHS::UpdateDriveBits()
{
	fDriveBits = 0;
	for (int i=0; i<8; i++) {
		if (fPartitionInfo[i].DriveNumber >= 1 && fPartitionInfo[i].DriveNumber <= 8
			&& fPartitionInfo[i].IsActive) {
			fDriveBits |= 1 << (fPartitionInfo[i].DriveNumber-1);
		}
	}
}

bool MyidePartitionInfoCHS::IsActivePartition(unsigned int partno) const
{
	if (!IsValidPartitionNumber(partno)) {
		return false;
	}
	return fPartitionInfo[partno-1].IsActive;
}

unsigned int MyidePartitionInfoCHS::GetPartitionSize(unsigned int num) const
{
	if (!IsActivePartition(num)) {
		return false;
	}
	return fPartitionInfo[num-1].MaxSectors;
}

EDensity MyidePartitionInfoCHS::GetPartitionDensity(unsigned int num) const
{
	if (!IsActivePartition(num)) {
		return eUnknownDensity;
	}
	return fPartitionInfo[num-1].Density;
}

bool MyidePartitionInfoCHS::IsValidPartitionSector(unsigned int partno, unsigned int secno) const
{
	if (!IsValidPartitionNumber(partno)) {
		Assert(false);
		return false;
	}
	return (secno >= 1) && (secno <= fPartitionInfo[partno-1].MaxSectors);
}

unsigned int MyidePartitionInfoCHS::GetPartitionSectorNumber(unsigned int partno, unsigned int sector) const
{
	return fPartitionInfo[partno-1].StartingCylinder * fSectorsPerCylinder
		+ sector - 1;
}

bool MyidePartitionInfoCHS::ImageSpaceEnabled() const
{
	return (fImageSpaceCylinder > 0) && (fImageSpaceCylinder < fCylinders) && (fCylindersPerImageSlot > 0);
}

unsigned int MyidePartitionInfoCHS::GetNumberOfImageSlots() const
{
	if (!ImageSpaceEnabled()) {
		return 0;
	}
	return (fCylinders - fImageSpaceCylinder) / fCylindersPerImageSlot;
}

unsigned int MyidePartitionInfoCHS::GetSectorsPerImageSlot() const
{
	if (!ImageSpaceEnabled()) {
		return 0;
	}
	return fCylindersPerImageSlot * fSectorsPerCylinder - 1;
}

bool MyidePartitionInfoCHS::IsValidImageSector(unsigned int secno) const
{
	// last sector of image slot is used for image infos
	return (secno >= 1) && (secno < fCylindersPerImageSlot * fSectorsPerCylinder);
}

unsigned int MyidePartitionInfoCHS::GetImageSectorNumber(unsigned int imageno, unsigned int sector) const
{
	return (fImageSpaceCylinder + (imageno - 1) * fCylindersPerImageSlot) * fSectorsPerCylinder
		+ sector - 1;
}

unsigned int MyidePartitionInfoCHS::GetImageInfoSectorNumber(unsigned int imageno) const
{
	// last sector of image slot
	return (fImageSpaceCylinder + imageno * fCylindersPerImageSlot) * fSectorsPerCylinder - 1;
}


bool MyidePartitionInfoCHS::IsValidSectorLength(unsigned int sector, EDensity density, unsigned int seclen) const
{
	if (density == eDD) {
		if (sector <= 3 && seclen != 128) {
			return false;
		}
		if (sector > 3 && seclen != 256) {
			return false;
		}
	} else {
		if (seclen != 128) {
			return false;
		}
	}
	return true;
}


