/*
   MyidePartitionInfoLBA - infos for (new) LBA partition table

   Copyright (C) 2008 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string.h>

#include "MyidePartitionInfoLBA.h"
#include "MiscUtils.h"

MyidePartitionInfoLBA::MyidePartitionInfoLBA()
	: fTotalLBASectors(0),
	  fImageSpaceStartLBA(0),
	  fImageSpaceSlots(0)
{ }

MyidePartitionInfoLBA::~MyidePartitionInfoLBA()
{ }

bool MyidePartitionInfoLBA::ParsePartitionInfo(const unsigned char* buf)
{
	if (buf[eOffsetDriveNumbersPartitionVersion] != eLBAPartitionTableMagic) {
		return false;
	}

	fTotalLBASectors = buf[eOfsTotalLBASize]
		+ (buf[eOfsTotalLBASize + 1] << 8)
		+ (buf[eOfsTotalLBASize + 2] << 16)
		+ (buf[eOfsTotalLBASize + 3] << 24);

	fDriveBits = buf[eOfsDriveBits];
	fOptionByte = buf[eOfsOptionByte];

	for (int i=0;i<eMaxPartitions;i++) {
		int idx = i*ePartitionBlockSize + ePartitionOffset;

		unsigned char driveno = buf[idx + eOfsPartitionDriveDensity];

		fPartitionInfo[i].DriveNumber = driveno & 0x0f;
		switch (driveno & 0xf0) {
		case ePartSD:
			fPartitionInfo[i].Density = eSD;
			break;
		case ePartDD:
			fPartitionInfo[i].Density = eDD;
			break;
		case ePartQD:
			fPartitionInfo[i].Density = eQD;
			break;
		default:
			fPartitionInfo[i].Density = eUnknownDensity;
		}
			
		fPartitionInfo[i].StartingLBA =
			 buf[idx + eOfsPartitionStart]
			 + (buf[idx + eOfsPartitionStart + 1] << 8)
			 + (buf[idx + eOfsPartitionStart + 2] << 16)
			 + (buf[idx + eOfsPartitionStart + 3] << 24);

		fPartitionInfo[i].MaxSectors =
			 buf[idx + eOfsPartitionSize]
			 + (buf[idx + eOfsPartitionSize + 1] << 8)
			 + (buf[idx + eOfsPartitionSize + 2] << 16);

		if (fPartitionInfo[i].DriveNumber >= 1 && fPartitionInfo[i].DriveNumber <=8) {
			fPartitionInfo[i].IsActive = DriveIncludedInDriveBits(fPartitionInfo[i].DriveNumber);
		} else {
			fPartitionInfo[i].IsActive = false;
		}
	}

	fImageSpaceStartLBA =
		buf[eImageSlotOffset + eOfsImageSlotStart]
		+ (buf[eImageSlotOffset + eOfsImageSlotStart + 1] << 8)
		+ (buf[eImageSlotOffset + eOfsImageSlotStart + 2] << 16)
		+ (buf[eImageSlotOffset + eOfsImageSlotStart + 3] << 24);

	fImageSpaceSlots =
		buf[eImageSlotOffset + eOfsImageSlotMax]
		+ (buf[eImageSlotOffset + eOfsImageSlotMax + 1] << 8);

	return true;
}

bool MyidePartitionInfoLBA::CreatePartitionInfo(unsigned char* buf) const
{
	memset(buf, 0, 256);

	// general stuff
	buf[eOfsTotalLBASize] = fTotalLBASectors & 0xff;
	buf[eOfsTotalLBASize + 1] = (fTotalLBASectors >> 8) & 0xff;
	buf[eOfsTotalLBASize + 2] = (fTotalLBASectors >> 16) & 0xff;
	buf[eOfsTotalLBASize + 3] = (fTotalLBASectors >> 24) & 0xff;

	buf[eOffsetDriveNumbersPartitionVersion] = eLBAPartitionTableMagic;

	buf[eOfsDriveBits] = fDriveBits;
	buf[eOfsOptionByte] = fOptionByte;

	// partition stuff
	for (int i=0;i<eMaxPartitions;i++) {
		int idx = i*ePartitionBlockSize + ePartitionOffset;

		unsigned char driveno = fPartitionInfo[i].DriveNumber;

		switch (fPartitionInfo[i].Density) {
		case eSD:
			driveno |= ePartSD;
			break;
		case eDD:
			driveno |= ePartDD;
			break;
		case eQD:
			driveno |= ePartQD;
			break;
		default:
			break;
		}
		buf[idx + eOfsPartitionDriveDensity] = driveno;

		buf[idx + eOfsPartitionStart] = fPartitionInfo[i].StartingLBA & 0xff;
		buf[idx + eOfsPartitionStart + 1] = (fPartitionInfo[i].StartingLBA >> 8) & 0xff;
		buf[idx + eOfsPartitionStart + 2] = (fPartitionInfo[i].StartingLBA >> 16) & 0xff;
		buf[idx + eOfsPartitionStart + 3] = (fPartitionInfo[i].StartingLBA >> 24) & 0xff;

		buf[idx + eOfsPartitionSize] = fPartitionInfo[i].MaxSectors & 0xff;
		buf[idx + eOfsPartitionSize + 1] = (fPartitionInfo[i].MaxSectors >> 8) & 0xff;
		buf[idx + eOfsPartitionSize + 2] = (fPartitionInfo[i].MaxSectors >> 24) & 0xff;

	}

	// image stuff

	buf[eImageSlotOffset + eOfsImageSlotStart] = fImageSpaceStartLBA & 0xff;
	buf[eImageSlotOffset + eOfsImageSlotStart + 1] = (fImageSpaceStartLBA >> 8) & 0xff;
	buf[eImageSlotOffset + eOfsImageSlotStart + 2] = (fImageSpaceStartLBA >> 16) & 0xff;
	buf[eImageSlotOffset + eOfsImageSlotStart + 3] = (fImageSpaceStartLBA >> 24) & 0xff;

	buf[eImageSlotOffset + eOfsImageSlotMax] = fImageSpaceSlots & 0xff;
	buf[eImageSlotOffset + eOfsImageSlotMax + 1] = (fImageSpaceSlots >> 8) & 0xff;

	return true;
}

bool MyidePartitionInfoLBA::IsLbaPartitionTable() const
{
	return true;
}

void MyidePartitionInfoLBA::PrintPartitionTable() const
{
	printf("\nMyIDE 5.x LBA partition table:\n");
        printf("total LBA sectors: %9d / $%07X\n", fTotalLBASectors, fTotalLBASectors);
        printf("Image space: ");
        if (ImageSpaceEnabled()) {
                printf("first sector: %9d / $%07X total slots: %d\n",
			fImageSpaceStartLBA,
			fImageSpaceStartLBA,
                        GetNumberOfImageSlots());
        } else {
                printf("disabled\n");
        }
	char bin[9];
	MiscUtils::byte_to_bin(fOptionByte, bin);
        printf("option byte: %s", bin);
	MiscUtils::byte_to_bin(fDriveBits, bin);
        printf(" drive bits: %s\n", bin);

        printf("part active drv          start          dns sectors\n");
        for (int i=1;i<=eMaxPartitions;i++) {
                printf("%d    %s    D%d:  %9d / $%07X   %s   %5d\n", i,
                        fPartitionInfo[i-1].IsActive ? "yes" : "no ",
                        fPartitionInfo[i-1].DriveNumber,
			fPartitionInfo[i-1].StartingLBA,
			fPartitionInfo[i-1].StartingLBA,
                        MyidePartitionInfo::DensityAsString(fPartitionInfo[i-1].Density),
                        fPartitionInfo[i-1].MaxSectors);
        }

}

void MyidePartitionInfoLBA::UpdateDriveBits()
{
	fDriveBits = 0;
	for (int i=0; i<8; i++) {
		if (fPartitionInfo[i].DriveNumber >= 1 && fPartitionInfo[i].DriveNumber <= 8
			&& fPartitionInfo[i].IsActive) {
			fDriveBits |= 1 << (fPartitionInfo[i].DriveNumber-1);
		}
	}
}


bool MyidePartitionInfoLBA::IsActivePartition(unsigned int partno) const
{
	if (!IsValidPartitionNumber(partno)) {
		return false;
	}
	return fPartitionInfo[partno-1].IsActive;
}

unsigned int MyidePartitionInfoLBA::GetPartitionSize(unsigned int num) const
{
	if (!IsActivePartition(num)) {
		return false;
	}
	return fPartitionInfo[num-1].MaxSectors;
}

EDensity MyidePartitionInfoLBA::GetPartitionDensity(unsigned int num) const
{
	if (!IsActivePartition(num)) {
		return eUnknownDensity;
	}
	return fPartitionInfo[num-1].Density;
}

bool MyidePartitionInfoLBA::IsValidPartitionSector(unsigned int partno, unsigned int secno) const
{
	if (!IsValidPartitionNumber(partno)) {
		Assert(false);
		return false;
	}
	return (secno >= 1) && (secno <= fPartitionInfo[partno-1].MaxSectors);
}

unsigned int MyidePartitionInfoLBA::GetPartitionSectorNumber(unsigned int partno, unsigned int sector) const
{
	return fPartitionInfo[partno-1].StartingLBA + sector;
}

// image stuff
bool MyidePartitionInfoLBA::ImageSpaceEnabled() const
{
	return (fImageSpaceStartLBA > 0) && (fImageSpaceSlots > 0);
}

unsigned int MyidePartitionInfoLBA::GetNumberOfImageSlots() const
{
	if (!ImageSpaceEnabled()) {
		return 0;
	}
	return fImageSpaceSlots;
}

unsigned int MyidePartitionInfoLBA::GetSectorsPerImageSlot() const
{
	if (!ImageSpaceEnabled()) {
		return 0;
	}
	// note: first sector is reserved for image info!
	return eSectorsPerImageSlot - 1;
}

bool MyidePartitionInfoLBA::IsValidImageSector(unsigned int secno) const
{
	// note: first sector is reserved for image info!
	return (secno >= 1) && (secno < eSectorsPerImageSlot);
}

unsigned int MyidePartitionInfoLBA::GetImageSectorNumber(unsigned int imageno, unsigned int sector) const
{
	return fImageSpaceStartLBA + (imageno - 1) * eSectorsPerImageSlot + sector;
}

unsigned int MyidePartitionInfoLBA::GetImageInfoSectorNumber(unsigned int imageno) const
{
	// first sector of image slot
	return fImageSpaceStartLBA + (imageno - 1) * eSectorsPerImageSlot;
}


// other stuff

bool MyidePartitionInfoLBA::IsValidSectorLength(unsigned int sector, EDensity density, unsigned int seclen) const
{
	if (density == eDD) {
		if (sector <= 3 && seclen != 128) {
			return false;
		}
		if (sector > 3 && seclen != 256) {
			return false;
		}
	} else {
		if (seclen != 128) {
			return false;
		}
	}
	return true;
}

bool MyidePartitionInfoLBA::SetPartition(unsigned int partition, bool active, unsigned int driveno, unsigned int start, unsigned int sectors, EDensity density)
{
	if (!IsValidPartitionNumber(partition)) {
		return false;
	}
	fPartitionInfo[partition-1].IsActive = active;
	fPartitionInfo[partition-1].DriveNumber = driveno;
	fPartitionInfo[partition-1].StartingLBA = start;
	fPartitionInfo[partition-1].MaxSectors = sectors;
	fPartitionInfo[partition-1].Density = density;
	UpdateDriveBits();
	return true;
}

bool MyidePartitionInfoLBA::SetImageSpace(unsigned int start, unsigned int num_slots)
{
	if (num_slots) {
		fImageSpaceStartLBA = start;
	} else {
		fImageSpaceStartLBA = 0;
	}
	fImageSpaceSlots = num_slots;
	return true;
}

void MyidePartitionInfoLBA::SetOptionByte(unsigned char byte)
{
	fOptionByte = byte;
}
