/*
   RawDiskWin32 - direct drive access for Win32

   Copyright (C) 2006-2008 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "RawDiskWin32.h"
#include "Error.h"
#include "AtariDebug.h"

#include <winioctl.h>

RawDiskWin32::RawDiskWin32(const char* filename)
	: super()
{
	fHandle = CreateFile(filename, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE , NULL, OPEN_EXISTING, 0, 0);
	if (fHandle == INVALID_HANDLE_VALUE) {
		throw FileOpenError(filename);
	}
}

RawDiskWin32::~RawDiskWin32()
{
	if (fHandle != INVALID_HANDLE_VALUE) {
		CloseHandle(fHandle);
		fHandle = INVALID_HANDLE_VALUE;
	}
}

bool RawDiskWin32::SeekToSector(unsigned int sector)
{
	if (fHandle == INVALID_HANDLE_VALUE) {
		return false;
	}
	LARGE_INTEGER largeint;
	largeint.QuadPart = ((LONGLONG) sector) * 512;

	if (!SetFilePointerEx(fHandle, largeint, NULL, FILE_BEGIN)) {
		DPRINTF("seeking to sector %d failed\n", sector);
		return false;
	}
	return true;
}

bool RawDiskWin32::ReadBlock(unsigned int sector, unsigned char* buf)
{
	if (!SeekToSector(sector)) {
		return false;
	}
	DWORD len;
	if (!ReadFile(fHandle, buf, 512, &len, NULL)) {
		DPRINTF("reading sector %d failed\n", sector);
		return false;
	}
	return true;
}

bool RawDiskWin32::WriteBlock(unsigned int sector, const unsigned char* buf)
{
	if (!SeekToSector(sector)) {
		return false;
	}
	DWORD len;
	if (!WriteFile(fHandle, buf, 512, &len, NULL)) {
		DPRINTF("writing sector %d failed\n", sector);
		return false;
	}
	return true;
}

unsigned int RawDiskWin32::GetDiskSizeInSectors() const
{
	if (fHandle == INVALID_HANDLE_VALUE) {
		return 0;
	}
	DWORD length;

	// note: the reported size may be lower than the real size when the drive
	// is connected via an USB bridge that translates cylinders/heads/sectors
	// unfortunately the DISK_GET_LENGTH_INFO an DISK_GET_GEOMETRY_EX ioctls
	// doesn't exist in Win2k, so we cannot use them...

	DISK_GEOMETRY geometry;

	if (!DeviceIoControl(fHandle, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0, &geometry, sizeof(geometry), &length, NULL)) {
		DWORD err = GetLastError();
		printf("getting size of disk drive failed: %d\n", (int) err);
		return 0;
	}

	return geometry.Cylinders.QuadPart * geometry.TracksPerCylinder * geometry.SectorsPerTrack;
}
